% Code for Example 4.1 in the manuscript:
% solve the "model problem" for a discretized
% 1-dim Laplacian with zero boundary conditions.

% The solution is computed in several different ways:
% 
% - v is computed via formula (2.7)
%   and structured linear system solving,
% - v_acc is computed via the accelerated formula (2.8)
%   and structured linear system solving,
% - v_exact is an accurate solution computed via expm 
%   and multiprecision (for comparison purposes).
%
% Note that vpa is used for computing the
% "exact" solution. If you do not have the
% Symbolic Toolbox, just comment/uncomment
% the relevant parts below.
%
% The script plots errors of v and v_acc wrt v_exact
% for an increasing number of terms, as in the manuscript.

% Choose parameters:

N=100;  % matrix size
t=pi/2; % point where the solution is evaluated,
gvec=rand(N,1); % this is vector g in the manuscript

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% L is our Laplacian:
L=toeplitz([2;-1;zeros(N-2,1)],[2,-1,zeros(1,N-2)]);
% but we will also need L^2, which will be called LL:
LL=L*L;

% Set up qs generators for LL.

d=[5;6*ones(N-2,1);5];
p=[[0 0];[ones(N-1,1),-2*ones(N-1,1)]];
h=[zeros(2,1),[-2*ones(1,N-1);ones(1,N-1)]];
a=zeros(2,2,N-1);
for k=1:N-1
    a(1,2,k)=1;
end
g=[ones(N-1,1),-2*ones(N-1,1)];
q=[-2*ones(1,N-1);ones(1,N-1)];
b=zeros(2,2,N-1);
for k=1:N-1
    b(1,2,k)=1;
end

% Check:
% LL_qs=recover_qs(p,q,a,g,h,b,d);
% should be the same as LL


% Let us compute the solution.

Ag=L*gvec; % (classical product, but might as well be done using matxvec.m)

% Compute "exact" solution with vpa:
 
 m1=expm(vpa(L)*vpa(2)*vpa(pi))-vpa(eye(N));
 m2=vpa(2)*vpa(pi)*vpa(L)*expm(vpa(L)*vpa(t));
 v_exact=double(vpa(m2*(m1\gvec)));

% ...or without vpa:
%
% m1=expm(L*2*pi)-eye(N);
% m2=2*pi*L*expm(L*t);
% v_exact=m2*(m1\gvec);


% Apply series expansion formulas (2.7) and (2.8) with structured method.
V0=1;
V1=t-pi;
V2=2*(pi^2/6-pi*t/2+t*t/4);
V3=2*(pi^2*t/6-pi*t^2/4+t^3/12);
A3g=L*L*Ag;
results=zeros(50,2);

% compute expansions of orders 10:10:500

v=gvec-(pi-t)*Ag;
v_acc=V0*gvec+V1*Ag+V2*L*Ag+V3*A3g;

for M=1:50 
  x=qs_shifted_function(p,q,a,g,h,b,d,(10*(M-1)+1:10*M).^2,Ag);
  x_acc=qs_shifted_function(p,q,a,g,h,b,d,(10*(M-1)+1:10*M).^2,A3g);
for k=1:10
    kk=10*(M-1)+k;
    v=v+2*(L*cos(kk*t)+LL*(sin(kk*t)/kk))*x(:,k);
    v_acc=v_acc-(2/kk^2)*(L*cos(kk*t)+LL*(sin(kk*t)/kk))*x_acc(:,k);
end

display(M)
results(M,:)=[norm(v-v_exact),norm(v_acc-v_exact)];
end

% Plot output:

figure(1)
title('Ex. 4.1')
xlabel('number of terms')
ylabel('log_{10}(error)')
hold on
plot(10:10:500,log10(results(:,1)),'bo-')
plot(10:10:500,log10(results(:,2)),'rd-')
hold off
legend('series expansion','accelerated expansion')
