% Code for Example 4.3 in the manuscript.
% 
% This example deals with the discretization of the
% Poisson equation on a rectangle of sides a and b.
% The discretization uses a grid of size Na x Nb
% and yields a matrix equation AX+XB=F.
%
% The solution is computed in two ways:
% - by solving the associated Kronecker linear system
%   with a classical solver,
% - by transforming the equation in a set of linear 
%   systems via Schur factorization and solving the 
%   systems with the structured method.
%
% The script outputs the absolute error on the
% solution, that is, the 2-norm distance between the
% two computed solutions.
%

% Choose Na and Nb.
Na=50;
Nb=50;

% Form coefficients of the matrix equation.
A=toeplitz([2;-1;zeros(Nb-2,1)],[2,-1,zeros(1,Nb-2)]);
B=toeplitz([2;-1;zeros(Na-2,1)],[2,-1,zeros(1,Na-2)]);
F=ones(Nb,Na);

% Kronecker approach:
C=kron(eye(Na),A)+kron(B,eye(Nb));
y=C\F(:);
Y=reshape(y,Nb,Na);

% Other approach: structured shifted systems.
[U,D]=schur(B);  % this is actually a well-known diagonalization
% now B=U*D*U' and we solve for XU
G=F*U; % new rhs

% Form quasiseparable generators for A:
d=2*ones(Nb,1);
p=[0;ones(Nb-1,1)];
q=-ones(1,Nb-1);
a=zeros(1,1,Nb-1);
g=-ones(Nb-1,1);
b=zeros(1,1,Nb-1);
h=[0,ones(1,Nb-1)];
% Check:
% A_qs=recover_qs(p,q,a,g,h,b,d);

% Solve the shifted systems and recover solution
x=qs_shifted_function_Sylvester(p,q,a,g,h,b,d,diag(D),G);
X=x*U';

fprintf('%s%10.2e\n','error =', norm(Y-X)/norm(Y))
